/*
    c99tree - input layer for line based text formats
    Copyright (C) 2010  Tibor Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "tree.h"
#include "tree_load.h"
#include "c99tree.h"
#include "escaping.h"

FILE *open_or_stdin(const char *fn)
{
	if (fn == NULL)
		return stdin;
	return fopen(fn, "r");
}

node_t *tree_load_ind(const char *fn)
{
	char *old_loc_format;
	FILE *f;
	char line[1024];
	char *type, *name, *loc;
	int level;
	int lineno;
	node_t *root = NULL;
	node_t *n;
	node_t *last[LOADER_MAXLEVEL];

	f = open_or_stdin(fn);
	if (f == NULL)
		return NULL;

	/* workaround for FILE node */
	lineno = 1;
	old_loc_format = loc_format;
	loc_format = NULL;
	fgets(line, sizeof(line), f);
	if (strncmp(line, "FILE", 4) != 0) {
		fprintf(stderr, "Error loading ind %s:%d: first node is not FILE\n", fn, lineno);
		goto err;
	}
	loc = strrchr(line, '(');
	if (loc == NULL) {
		fprintf(stderr, "Error loading ind %s:%d: FILE node doesn't have location format\n", fn, lineno);
		goto err;
	}
	loc[-1] = '\0';
	loc++;
	loc[strlen(loc)-2] = '\0';
	loc_format = strdup(loc);

	while(!(feof(f))) {
		/* read line */
		lineno++;
		*line = '\0';
		fgets(line, sizeof(line), f);

		if ((*line == '\0') || (*line == '\n'))
			continue;

		/* split in 3 fields and tell level */
		type = line;
		level = -1; /* workaround for FILE node */
		while(*type == ' ') {
			type++;
			level++;
		}
		if (level >= LOADER_MAXLEVEL) {
			fprintf(stderr, "Error loading ind %s:%d: nesting too deep(%d)\n", fn, lineno, level);
			goto err;
		}
		name = strchr(type, ' ');
		if (name == NULL) {
			fprintf(stderr, "Error loading ind %s:%d: no name\n", fn, lineno);
			goto err;
		}
		*name = '\0';
		name++;
		loc = strrchr(name, '(');
		if (loc == NULL) {
			fprintf(stderr, "Error loading ind %s:%d: no location\n", fn, lineno);
			goto err;
		}

		/* cut at space before opening '(', truncate closing ')' */
		loc[-1] = '\0';
		loc++;
		loc[strlen(loc)-2] = '\0';

		/* create node */
		loc_current = loc_scan(loc);
		if (loc_current == NULL) {
			fprintf(stderr, "Error loading ind %s:%d: bad location '%s'\n", fn, lineno, loc);
			goto err;
		}
		n = Nalloc(Ntype(type), unescape_dup(name), NULL);

		/* bind new node */
		if (level == 0) {
			if (root == NULL)
				root = n;
			else
				Nappend(root, n);
		}
		else {
			Nchappend(last[level-1], n);
		}
		last[level] = n;
	}
	free(loc_format);
	loc_format = old_loc_format;
	loc_current = NULL;
	fclose(f);
	return root;
err:;
	if (loc_format != NULL)
		free(loc_format);
	loc_format = old_loc_format;
	loc_current = NULL;
	fclose(f);
	if (root != NULL)
		Nfree(root);
	return NULL;
}
